from dataclasses import dataclass, field
from typing import Optional

from idp_authentication.custom_types import (
    AppEntitiesRestrictionsDict,
    PermissionRestrictionsDict,
)
from idp_authentication.users.base_classes.base_entity import BaseEntity


@dataclass(kw_only=True)
class UserRole(BaseEntity):
    role: str = field(default_factory=str)
    app_entities_restrictions: AppEntitiesRestrictionsDict = field(default_factory=dict)
    permission_restrictions: PermissionRestrictionsDict = field(default_factory=dict)
    user_id: Optional[str] = field(default=None)



from typing import Any

from idp_authentication.enums import ChoiceEnum
from idp_authentication.users.base_classes.base_use_case import BaseUseCase


class GetUsersWithAccessToAppEntityRecordUseCase(BaseUseCase):
    def execute(
        self,
        app_entity_type: str,
        record_identifier: Any,
        roles: list[ChoiceEnum | str],
    ):
        self._verify_app_entity_type(app_entity_type)
        for role in roles:
            self._verify_role(role)

        return self.users_unit_of_work.user_repository.get_users_with_access_to_records(
            app_entity_type=app_entity_type,
            record_identifier=record_identifier,
            roles=roles,
        )



import uuid
from dataclasses import dataclass, field

import pytest as pytest
from dependency_injector import containers, providers
from faker import Faker
from sqlalchemy import Column, String, create_engine, orm
from sqlalchemy.orm import Session, mapper
from sqlalchemy.pool import NullPool

from idp_authentication.config import Config, get_test_memory_db_uri
from idp_authentication.users.base_classes.singleton import Singleton
from idp_authentication.users.di.containers import UsersModuleDIContainer
from idp_authentication.users.domain.entities import AppEntity, User, UserRole
from idp_authentication.users.infrastructure.database.sqlalchemy.base import (
    Base,
    BaseModelMixin,
)
from idp_authentication.users.infrastructure.database.sqlalchemy.custom_orm_column_types import (
    UUID,
)
from idp_authentication.users.tests.mock_event_producer import MockEventProducer

fake = Faker()

TEST_ROLE_1 = "TestRole_1"
TEST_ROLE_2 = "TestRole_2"
VEHICLE_APP_ENTITY_IDENTIFIER = "vehicle"
TEST_PERMISSION = "TestPermission"


class TestConfig(Config):
    APP_IDENTIFIER = "test"
    TENANTS = ["default"]
    ROLES = [TEST_ROLE_1, TEST_ROLE_2]
    APP_ENTITY_TYPES = [VEHICLE_APP_ENTITY_IDENTIFIER]

    class Config:
        env_file_encoding = "utf-8"
        use_enum_values = True


@dataclass(kw_only=True)
class Vehicle(AppEntity):
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    identifier: str = field(default_factory=str)
    label: str = field(default_factory=str)

    @property
    def idp_identifier(self):
        return self.identifier

    @property
    def idp_label(self):
        return self.label


class VehicleModel(Base, BaseModelMixin):
    __tablename__ = "vehicles"
    __table_args__ = {"extend_existing": True}

    id = Column(UUID(), primary_key=True, unique=True)
    identifier = Column(String(100), nullable=False, unique=True)
    label = Column(String(100), nullable=False)


def start_test_mappers():
    mapper(
        Vehicle,
        VehicleModel.__table__,
    )


class Database(metaclass=Singleton):
    def __init__(
        self, db_url: str, isolation_level: str = "READ COMMITTED", pool_class=NullPool
    ) -> None:
        self.db_url = db_url
        self._engine = create_engine(db_url, isolation_level=isolation_level)
        self._session_factory = orm.scoped_session(
            orm.sessionmaker(
                expire_on_commit=False,
                autoflush=False,
                bind=self._engine,
                info={"debug": True},
            ),
        )

    def create_database(self) -> None:
        Base.metadata.create_all(self._engine)

    def drop_all(self) -> None:
        if "memory" in self.db_url:
            Base.metadata.drop_all(self._engine)

    @property
    def session(self) -> Session:
        return self._session_factory()


class TestContainer(containers.DeclarativeContainer):
    config = providers.Configuration(pydantic_settings=[TestConfig()])
    url = get_test_memory_db_uri()
    database = providers.Singleton(
        Database,
        db_url=url,
        isolation_level=config.ISOLATION_LEVEL,
    )
    event_producer = providers.Factory(MockEventProducer)
    users_module = providers.Container(
        UsersModuleDIContainer,
        config=config,
        database=database,
        event_producer=event_producer,
    )

    @staticmethod
    def __new__(cls, *args, **kwargs):
        container = super().__new__(cls, *args, **kwargs)
        container.users_module.start_mappers()
        return container


@pytest.fixture(scope="session")
def container():
    container = TestContainer()
    container.wire(packages=["idp_authentication"])
    start_test_mappers()
    database = container.database()
    database.create_database()
    insert_initial_data(container)

    yield container
    container.reset_override()
    database.drop_all()


def insert_initial_data(container):
    session = container.database().session

    user_1 = User(
        username="test1",
        email="test@test.com",
        first_name="test",
        last_name="test",
        is_active=True,
        is_superuser=False,
        is_staff=False,
        user_roles=[
            UserRole(
                role=TEST_ROLE_1,
                app_entities_restrictions={VEHICLE_APP_ENTITY_IDENTIFIER: [1, 2]},
                permission_restrictions={
                    TEST_PERMISSION: {VEHICLE_APP_ENTITY_IDENTIFIER: [1]}
                },
            )
        ],
    )
    user_2 = User(
        username="test2",
        email="test@test2.com",
        first_name="test2",
        last_name="test2",
        is_active=True,
        is_superuser=False,
        is_staff=False,
        user_roles=[
            UserRole(
                role=TEST_ROLE_2,
                app_entities_restrictions={VEHICLE_APP_ENTITY_IDENTIFIER: [3]},
                permission_restrictions={
                    TEST_PERMISSION: {VEHICLE_APP_ENTITY_IDENTIFIER: [4]}
                },
            )
        ],
    )
    user_3 = User(
        username="test3",
        email="test@test3.com",
        first_name="test3",
        last_name="test3",
        is_active=True,
        is_superuser=False,
        is_staff=False,
    )
    vehicle = Vehicle(identifier="1", label="vehicle1")
    vehicle_2 = Vehicle(identifier="2", label="vehicle2")
    objects = [user_1, user_2, user_3, vehicle, vehicle_2]
    session.add_all(objects)
    session.commit()


@pytest.fixture(scope="function")
def user(container):
    session = container.database().session
    return session.query(User).filter(User.username == "test1").first()


@pytest.fixture(scope="function")
def user_2(container):
    session = container.database().session
    return session.query(User).filter(User.username == "test2").first()


@pytest.fixture(scope="function")
def user_without_role(container):
    session = container.database().session
    return session.query(User).filter(User.username == "test3").first()


@pytest.fixture(scope="function")
def make_user(container):
    def create_user(**overrides):
        default_data = {
            "username": fake.user_name(),
            "email": fake.email(),
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
            "is_active": True,
            "is_superuser": False,
            "is_staff": False,
            "user_roles": [
                UserRole(
                    role=TEST_ROLE_1,
                    app_entities_restrictions={VEHICLE_APP_ENTITY_IDENTIFIER: [1, 2]},
                    permission_restrictions={
                        TEST_PERMISSION: {VEHICLE_APP_ENTITY_IDENTIFIER: [1]}
                    },
                )
            ],
        }
        default_data.update(overrides)

        user = User(**default_data)

        session = container.database.provided().session
        session.add(user)
        session.commit()

        return user

    return create_user


@pytest.fixture(scope="function")
def vehicle(container):
    session = container.database().session
    return session.query(Vehicle).first()



from abc import ABC, abstractmethod


class ExceptionBase(Exception, ABC):
    def __init__(self, message, metadata=None):
        super().__init__(message)
        self.message = message
        self.metadata = metadata

    def __str__(self):
        return self.message

    @property
    @abstractmethod
    def code(self) -> str:
        """code must be implemented by subclass"""

    def to_dictionary(self) -> dict:
        return {"message": self.message, "code": self.code, "metadata": self.metadata}



import pytest

from idp_authentication.conftest import (
    TEST_PERMISSION,
    TEST_ROLE_1,
    VEHICLE_APP_ENTITY_IDENTIFIER,
)
from idp_authentication.exceptions import AccessDenied


@pytest.mark.parametrize(
    "app_entity_records_identifiers, permission",
    [
        ([1], None),
        ([1, 2], None),
        ([1], TEST_PERMISSION),
    ],
)
def test_authorize_app_entity_records_does_not_raise_error_on_valid_access(
    container, user, app_entity_records_identifiers, permission
):
    authorize_app_entity_records = (
        container.users_module().use_cases().authorize_app_entity_records_use_case()
    )
    authorize_app_entity_records.execute(
        user=user,
        role=TEST_ROLE_1,
        app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
        app_entity_records_identifiers=app_entity_records_identifiers,
        permission=permission,
    )


@pytest.mark.parametrize(
    "app_entity_records_identifiers, permission",
    [
        ([3], None),
        ([2], TEST_PERMISSION),
    ],
)
def test_authorize_app_entity_records_raises_error_on_invalid_access(
    container, user, app_entity_records_identifiers, permission
):
    authorize_app_entity_records = (
        container.users_module().use_cases().authorize_app_entity_records_use_case()
    )
    with pytest.raises(
        AccessDenied,
        match="You are not allowed to access the records in the requested entity",
    ):
        authorize_app_entity_records.execute(
            user=user,
            role=TEST_ROLE_1,
            app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
            app_entity_records_identifiers=app_entity_records_identifiers,
            permission=permission,
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass
class AppEntity(ABC):
    """Base class for all entities that are part of an application."""

    @property
    @abstractmethod
    def idp_identifier(self):
        """Returns the identifier attribute of the app entity."""

    @property
    @abstractmethod
    def idp_label(self):
        """Returns the label attribute of the app entity."""

    @property
    def entity_type(self):
        """
        Returns the entity type of the app entity.
        By default, this is the class name of the app entity, to lower.
        Can be overridden in subclasses.
        """
        return self.__class__.__name__.lower()



from idp_authentication.users.adapters.repositories.user_repository import (
    UserRepository,
)
from idp_authentication.users.adapters.repositories.user_role_repository import (
    UserRoleRepository,
)
from idp_authentication.users.domain.ports.session import SessionPort
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort


class UsersUnitOfWork(UsersUnitOfWorkPort):
    def __init__(self, session: SessionPort):

        self.session = session
        self.user_repository = UserRepository(session)
        self.user_role_repository = UserRoleRepository(session)



from typing import Any, Optional

from idp_authentication.enums import ChoiceEnum
from idp_authentication.users.base_classes.base_use_case import BaseUseCase, UseCasePort
from idp_authentication.users.domain.ports import unit_of_work


class AuthorizeOrGetAllAllowedRecordIdentifiersUseCase(BaseUseCase):
    def __init__(
        self,
        users_unit_of_work: unit_of_work.UsersUnitOfWorkPort,
        roles: ChoiceEnum,
        app_entity_types: list[str],
        authorize_app_entity_records_use_case: UseCasePort,
        get_allowed_record_identifiers_use_case: UseCasePort,
    ):
        super().__init__(
            users_unit_of_work=users_unit_of_work,
            roles=roles,
            app_entity_types=app_entity_types,
        )
        self.authorize_app_entity_records_use_case = (
            authorize_app_entity_records_use_case
        )
        self.get_allowed_record_identifiers_use_case = (
            get_allowed_record_identifiers_use_case
        )

    def execute(
        self,
        user,
        role: ChoiceEnum,
        app_entity_type: str,
        app_entity_records_identifiers: Optional[list[Any]],
        permission: str = None,
    ) -> list[Any]:
        """
        Make sure that the user can access the entity records being requested and return them.
        If no identifiers are being provided, return all allowed entity records.
        Args:
            user:                           The user performing the request
            role:                           The role that the user is acting as.
            app_entity_type:                The app entity being accessed
            app_entity_records_identifiers: The identifiers of the records belonging to the specified app entity
            permission:                     In case of specific permissions we can have permission restrictions
                                                through IDP. The value is the name of the permission
        Raises:
            PermissionDenied: In case the requested records are not allowed
        """

        if app_entity_records_identifiers:
            self.authorize_app_entity_records_use_case.execute(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
                app_entity_records_identifiers=app_entity_records_identifiers,
            )
            return app_entity_records_identifiers
        else:
            return self.get_allowed_record_identifiers_use_case.execute(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
